package comunicacao;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;

import comunicacao.data.segment.Segment;
import comunicacao.util.ConnectionState;

public class ServerSocket implements IServerSocket {
	private DatagramSocket socket;
	private byte[] buffer;

	public ServerSocket(int localPort) throws SocketException {
		this.socket = new DatagramSocket(localPort);
		this.buffer = new byte[1024];
	}

	@Override
	public ISocket accept() throws IOException {
		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		this.socket.receive(packet);
		Segment seg = new Segment(packet);
		Socket returnMe = null;
		if (seg.getHeader().isSYN()){
			returnMe = new Socket(packet.getAddress(), packet.getPort(), seg);
		}
		
		synchronized (returnMe.getStateMonitor()) {
			while (!returnMe.getState().equals(ConnectionState.OPEN)){
				try {
					returnMe.getStateMonitor().wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		
		
		

		return returnMe;
	}

	@Override
	public int getLocalPort() {
		return this.socket.getLocalPort();
	}

	@Override
	public InetAddress getLocalAddress() {
		return this.socket.getLocalAddress();
	}

}
