package comunicacao;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import comunicacao.data.SegmentsConstructor;
import comunicacao.data.segment.Header;
import comunicacao.data.segment.HeaderFlags;
import comunicacao.data.segment.Segment;
import comunicacao.data.segment.SequenceNumber;
import comunicacao.exception.HeaderDontFitException;
import comunicacao.exception.NotConnectedException;
import comunicacao.sender.ReceiverThread;
import comunicacao.sender.SendQueue;
import comunicacao.util.ConnectionState;

public class Socket implements ISocket {
	//TODO Check synchronization of send and receive variables EVERYWHERE
	//TODO Check .clone() when creating new segments
	//TODO Modify the sets in Segment to make them alter the header inside the packet.
	
	

	private DatagramSocket socket;

	private InetAddress remoteAddress;
	private int remotePort;

	/*-------------------SENDING VARIABLES----------------*/
	private SequenceNumber sendInitialSeqNum;
	private SequenceNumber sendNextSeqNum;
	private SequenceNumber sendUnackSeqNum;
	private int sendWindow;
	private Object seqNumMonitor;

	/*-------------------RECEIVING VARIABLES--------------*/
	private SequenceNumber receiveInitialSeqNum;
	private SequenceNumber receiveNextSeqNum;
	private int receiveWindow;
	private Object receiveSeqNumMonitor;

	private ConnectionState state;

	private Random random;

	private SendQueue sender;
	private ReceiverThread receiver;

	private Object stateMonitor;
	private Object sendingMonitor;
	private Object receivingMonitor;

	private boolean activePassive;

	public Socket(InetAddress destinationIP, int destinationPort,
			boolean activePassive) throws SocketException {
		this.socket = new DatagramSocket();
		System.out.println(socket.getLocalPort());
		
		this.remoteAddress = destinationIP;
		this.remotePort = destinationPort;
		
		this.sendInitialSeqNum = null;
		this.sendNextSeqNum = null;
		this.sendUnackSeqNum = null;
		this.sendWindow = Segment.MAXIMUM_SEGMENT_SIZE <<4;
		this.seqNumMonitor = new Object();
		
		this.receiveInitialSeqNum = null;
		this.receiveNextSeqNum = null;
		this.receiveWindow = Segment.MAXIMUM_SEGMENT_SIZE <<3;
		this.receiveSeqNumMonitor = new Object();
		
		this.state = ConnectionState.CLOSED;
		
		this.random = new Random();
		
		this.sender = new SendQueue(this);
		this.receiver = new ReceiverThread(this);
		
		this.stateMonitor = new Object();
		this.sendingMonitor = new Object();
		this.receivingMonitor = new Object();
		
		this.activePassive = activePassive;
	

		this.sender = new SendQueue(this);
		this.receiver = new ReceiverThread(this);
		Thread t = new Thread(this.sender);
		//t.setDaemon(true);
		t.start();
		t = new Thread(this.receiver);
		//t.setDaemon(true);
		t.start();
		

		if (this.activePassive) { // if the socket is "active"
			this.open(destinationIP, destinationPort);
		} else { // if the socket is "passive"
			this.setState(ConnectionState.OPENING);
		}

	}	
	
	public Socket(int localPort) throws SocketException {
		this.socket = new DatagramSocket(localPort);
		
		this.remoteAddress = null;
		this.remotePort = 0;
		
		this.sendInitialSeqNum = null;
		this.sendNextSeqNum = null;
		this.sendUnackSeqNum = null;
		this.sendWindow = Segment.MAXIMUM_SEGMENT_SIZE <<4;
		this.seqNumMonitor = new Object();
		
		this.receiveInitialSeqNum = null;
		this.receiveNextSeqNum = null;
		this.receiveWindow = Segment.MAXIMUM_SEGMENT_SIZE <<3;
		this.receiveSeqNumMonitor = new Object();
		
		this.state = ConnectionState.CLOSED;
		
		this.random = new Random();
		
		this.sender = new SendQueue(this);
		this.receiver = new ReceiverThread(this);
		
		this.stateMonitor = new Object();
		this.sendingMonitor = new Object();
		this.receivingMonitor = new Object();
		
		this.activePassive = false;
	

		this.sender = new SendQueue(this);
		this.receiver = new ReceiverThread(this);
		Thread t = new Thread(this.sender, "Sender");
		//t.setDaemon(true);
		t.start();
		t = new Thread(this.receiver, "Receiver");
		//t.setDaemon(true);
		t.start();
		
		this.setState(ConnectionState.OPENING);

	}
	
	public Socket (InetAddress remoteAddress, int remotePort, Segment syn) throws SocketException{
		this.socket = new DatagramSocket();
		this.remoteAddress = remoteAddress;
		this.remotePort = remotePort;
		this.random  = new Random();
		this.sendInitialSeqNum = new SequenceNumber(this.random.nextInt());
		this.sendNextSeqNum = new SequenceNumber(this.sendInitialSeqNum.getInteger() + 1);
		this.sendUnackSeqNum = this.sendInitialSeqNum.clone();
		this.sendWindow = Segment.MAXIMUM_SEGMENT_SIZE << 3;
		this.seqNumMonitor = new Object();
		this.receiveInitialSeqNum = syn.getSequenceNumber();
		this.receiveNextSeqNum = new SequenceNumber(this.receiveInitialSeqNum.getInteger() +1);
		this.receiveWindow = Segment.MAXIMUM_SEGMENT_SIZE <<3;
		this.receiveSeqNumMonitor = new Object();
		this.state = ConnectionState.OPENING;
		this.sender = new SendQueue(this);
		this.receiver = new ReceiverThread(this);
		this.stateMonitor = new Object();
		this.sendingMonitor = new Object();
		this.receivingMonitor = new Object();
		this.activePassive = true;
		
		new Thread(sender).start();
		new Thread(receiver).start();
		
		
		Header header = new Header(this.sendInitialSeqNum, this.receiveNextSeqNum.clone(), HeaderFlags.SYNACK);
		this.sender.submit(header);
		
		
		
		
	}
	
	
	public void open(InetAddress remoteAddress, int remotePort) {
		if (remoteAddress == null){
			throw new NullPointerException("remote address can't be null");
		}
		this.setState(ConnectionState.OPENING);
		this.remoteAddress = remoteAddress;
		this.remotePort = remotePort;
		
		this.sendInitialSeqNum = new SequenceNumber(this.random.nextInt());
		this.sendNextSeqNum = new SequenceNumber(
				this.sendInitialSeqNum.getInteger() + 1);
		this.sendUnackSeqNum = new SequenceNumber(
				this.sendInitialSeqNum.getInteger());

		Header header = new Header(this.sendInitialSeqNum, null, new HeaderFlags(true, false, false, false,
				false, false, 0));
		sender.submit(header);

		synchronized (stateMonitor) {
			while (!this.state.equals(ConnectionState.OPEN)) {
				try {
					stateMonitor.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		
//		try {
//			this.remotePort = (Integer) this.readObject();
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		} catch (ClassNotFoundException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		} catch (NotConnectedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		
	}
	
	

	public Socket(DatagramSocket socket, InetAddress remoteAddress,
			int remotePort, SequenceNumber sendInitialSeqNum,
			SequenceNumber sendNextSeqNum, SequenceNumber sendUnackSeqNum,
			int sendWindow, Object seqNumMonitor,
			SequenceNumber receiveInitialSeqNum,
			SequenceNumber receiveNextSeqNum, int receiveWindow,
			Object receiveSeqNumMonitor, ConnectionState state, Random random,
			SendQueue sender, ReceiverThread receiver, Object stateMonitor,
			Object sendingMonitor, Object receivingMonitor,
			boolean activePassive) {
		super();
		this.socket = socket;
		this.remoteAddress = remoteAddress;
		this.remotePort = remotePort;
		this.sendInitialSeqNum = sendInitialSeqNum;
		this.sendNextSeqNum = sendNextSeqNum;
		this.sendUnackSeqNum = sendUnackSeqNum;
		this.sendWindow = sendWindow;
		this.seqNumMonitor = seqNumMonitor;
		this.receiveInitialSeqNum = receiveInitialSeqNum;
		this.receiveNextSeqNum = receiveNextSeqNum;
		this.receiveWindow = receiveWindow;
		this.receiveSeqNumMonitor = receiveSeqNumMonitor;
		this.state = state;
		this.random = random;
		this.sender = sender;
		this.receiver = receiver;
		this.stateMonitor = stateMonitor;
		this.sendingMonitor = sendingMonitor;
		this.receivingMonitor = receivingMonitor;
		this.activePassive = activePassive;
	}

	public Socket(InetAddress byName, int i) throws SocketException {
		this(byName, i, true);
	}

	public ReceiverThread getReceiver() {
		return receiver;
	}

	public void setReceiver(ReceiverThread receiver) {
		this.receiver = receiver;
	}

	public Object getSendingMonitor() {
		return sendingMonitor;
	}

	public void setSendingMonitor(Object sendingMonitor) {
		this.sendingMonitor = sendingMonitor;
	}

	public boolean isActivePassive() {
		return activePassive;
	}

	public void setActivePassive(boolean activePassive) {
		this.activePassive = activePassive;
	}

	public void send(byte[] data, int offset, int length)
			throws NotConnectedException {

		if (this.state.equals(ConnectionState.OPEN)) {
			try {
				synchronized (sendingMonitor) {
					this.sender.submit(data, offset, length, sendingMonitor);
					// SegmentsConstructor segment = new
					// SegmentsConstructor(data, length, this.remoteAddress,
					// this.remotePort);
					sendingMonitor.wait();
				}
				// } catch (HeaderDontFitException e) {
				// e.printStackTrace();
			} catch (InterruptedException e) {
			}
		} else {
			throw new NotConnectedException();
		}
	}

	public void receive(byte[] data, int offset, int length)
			throws NotConnectedException {

		if (this.state.equals(ConnectionState.OPEN)) {
				this.receiver.read(data, offset, length);
		} else {
			throw new NotConnectedException();
		}
	}

	public void close() {
		this.setState(ConnectionState.CLOSING);
		Header header = new Header( this.sendNextSeqNum.clone(), null, new HeaderFlags(false, true, false, false,
				false, false, 0));
		this.sender.submit(header);
		

		synchronized (stateMonitor) {
			while (!this.state.equals(ConnectionState.CLOSED)) {
				try {
					stateMonitor.wait();
				} catch (InterruptedException e) {
				}
			}
		}

	}

	public int getPort() {
		return this.socket.getLocalPort();
	}

	public double getUploadRate() {
		if (this.sender != null) {
			return this.sender.getUploadRate();
		}

		return 0;
	}

	public double getDownloadRate() {
		if (this.receiver != null) {
			return this.receiver.getDownloadRate();
		}

		return 0;
	}

	@Override
	public void sendObject(Serializable obj) throws IOException, NotConnectedException {

		ByteArrayOutputStream arrayOut = new ByteArrayOutputStream();
		ObjectOutputStream out = new ObjectOutputStream(arrayOut);
		out.writeObject(obj);
		byte[] bytes = arrayOut.toByteArray();
		out.close();
		byte[] length = new byte[4];  
		length[3] = (byte) (bytes.length & 0xff);
		length[2] = (byte) ((bytes.length >> 8) & 0xff);
		length[1] = (byte) ((bytes.length >> 16) & 0xff);
		length[0] = (byte) ((bytes.length >> 24) & 0xff);

		
		
		this.send(length, 0, 4);	

		this.send(bytes, 0, bytes.length);
	
	}

	@Override
	public void sendFile(File file) throws IOException, NotConnectedException {

		byte[] buffer = new byte[5000000];
		FileInputStream fileIn = new FileInputStream(file);
		long fileLength = file.length();
		long aux = 0;
		int read;
		String path = file.getAbsolutePath();
		String fileName = path.substring(path.lastIndexOf('.'));

		this.sendObject(fileLength);
		this.sendObject(fileName);

		
		while (aux != fileLength) {
			read = fileIn.read(buffer);
			aux += read;
			this.send(buffer, 0, read);
			System.out.println("Total: " + fileLength + " Sent: " + aux);
		}
		fileIn.close();
		

	}

	@Override
	public Serializable readObject() throws IOException, ClassNotFoundException, NotConnectedException {
		
		
		byte[] lengthArray = new byte[4];
		this.receive(lengthArray, 0, 4);
		
		int length = lengthArray[0];
		length = (length << 8) + (lengthArray[1] & 0xff);
		length = (length << 8) + (lengthArray[2] & 0xff);
		length = (length << 8) + (lengthArray[3] & 0xff);

		
		byte[] object = new byte[length];

		this.receive(object, 0, length);
	
		ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(
				object));
		Serializable ser = (Serializable) in.readObject();
		in.close();
		
		return ser;
	}

	@Override
	public File readFile() throws IOException, ClassNotFoundException, NotConnectedException {
			FileOutputStream fileOut = null;
			File file;
			long fileLength;
			String fileName;
			int aux = 0, read = 0;
			byte[] buffer = new byte[5000000];
			
			(new File("temp")).mkdir();
	
			do {
				file = new File("temp\\" + random.nextInt(Integer.MAX_VALUE) + ".yamtemp");
			} while (file.exists());
	
			try {
				fileOut = new FileOutputStream(file, true);
			} catch (FileNotFoundException e) {
				// Won't happen
				e.printStackTrace();
			}
	
			fileLength = (Long) this.readObject();
	
			fileName = (String) this.readObject();
			// System.out.println(fileName);
			System.out.println("FILELENGTH " + fileLength);
			while (aux != fileLength) {
				if (fileLength - aux < buffer.length) {
					
					read = (int) (fileLength - aux);
				} else {
					read = buffer.length;
				}
				System.out.println("READ: " + read);
				
				this.receive(buffer, 0, read);
				
				aux += read;
				System.out.println("Total: " + fileLength + " Received: " + aux);
				fileOut.write(buffer, 0, read);
			}
	
			fileOut.close();
	
			File definite = new File("temp\\" +file.getName().substring(0, file.getName().lastIndexOf('.'))+ fileName);
			file.renameTo(definite);
			return definite;
	}

	

	

	@Override
	public void sendFile(File file, Object fileState) {
		// TODO  sendFile(File file, Object fileState)

	}

	public DatagramSocket getDatagramSocket() {
		return this.socket;
	}

	@Override
	public boolean connect() throws IOException {
		// TODO connect()
		return false;
	}

	public Random getRandom() {
		return random;
	}

	public void setRandom(Random random) {
		this.random = random;
	}

	public SequenceNumber getReceiveInitialSeqNum() {
		return receiveInitialSeqNum;
	}

	public void setReceiveInitialSeqNum(SequenceNumber receiveInitialSeqNum) {
		this.receiveInitialSeqNum = receiveInitialSeqNum;
	}

	public InetAddress getRemoteAddress() {
		return remoteAddress;
	}

	public void setRemoteAddress(InetAddress remoteAddress) {
		this.remoteAddress = remoteAddress;
	}

	public int getRemotePort() {
		return remotePort;
	}

	public void setRemotePort(int remotePort) {
		this.remotePort = remotePort;
	}

	public Object getStateMonitor() {
		return stateMonitor;
	}

	public void setStateMonitor(Object stateMonitor) {
		this.stateMonitor = stateMonitor;
	}

	public ConnectionState getState() {
		return state;
	}

	public void setState(ConnectionState state) {
		synchronized (stateMonitor) {
			System.out.println("Socket changing state to " + state.toString());
			this.state = state;
			stateMonitor.notifyAll();
		}		
	}

	public DatagramSocket getSocket() {
		return socket;
	}

	public void setSocket(DatagramSocket socket) {
		this.socket = socket;
	}

	public Object getReceiveSeqNumMonitor() {
		return receiveSeqNumMonitor;
	}

	public void setReceiveSeqNumMonitor(Object receiveSeqNumMonitor) {
		this.receiveSeqNumMonitor = receiveSeqNumMonitor;
	}

	public SequenceNumber getSendNextSeqNum() {
		return sendNextSeqNum;
	}

	public void setSendNextSeqNum(SequenceNumber sendNextSeqNum) {
		this.sendNextSeqNum = sendNextSeqNum;
	}

	public SendQueue getSender() {
		return sender;
	}

	public void setSender(SendQueue sender) {
		this.sender = sender;
	}

	public SequenceNumber getSendInitialSeqNum() {
		return sendInitialSeqNum;
	}

	public void setSendInitialSeqNum(SequenceNumber sendInitialSeqNum) {
		this.sendInitialSeqNum = sendInitialSeqNum;
	}

	public SequenceNumber getSendUnackSeqNum() {
		return sendUnackSeqNum;
	}

	public void setSendUnackSeqNum(SequenceNumber sendUnackSeqNum) {
		this.sendUnackSeqNum = sendUnackSeqNum;
	}

	public int getSendWindow() {
		return sendWindow;
	}

	public void setSendWindow(int sendWindow) {
		this.sendWindow = sendWindow;
	}

	public Object getSeqNumMonitor() {
		return seqNumMonitor;
	}

	public void setSeqNumMonitor(Object seqNumMonitor) {
		this.seqNumMonitor = seqNumMonitor;
	}

	public SequenceNumber getReceiveNextSeqNum() {
		return receiveNextSeqNum;
	}

	public void setReceiveNextSeqNum(SequenceNumber receiveNextSeqNum) {
		this.receiveNextSeqNum = receiveNextSeqNum;
	}

	public int getReceiveWindow() {
		return receiveWindow;
	}

	public void setReceiveWindow(int receiveWindow) {
		this.receiveWindow = receiveWindow;
	}

	public Object getReceivingMonitor() {
		return receivingMonitor;
	}

	public void setReceivingMonitor(Object receivingMonitor) {
		this.receivingMonitor = receivingMonitor;
	}

	@Override
	public double getPercentage() {
		// TODO Auto-generated method stub
		return 0;
	}

}
