package comunicacao.data.segment;

import java.net.DatagramPacket;

import comunicacao.exception.HeaderDontFitException;

/**
 * @author Joo Pedro
 *
 */
public class Segment {
	//private int segmentLength;
	//private int segmentWindow;
	
	public static final int MAXIMUM_SEGMENT_SIZE = 500;
	public static final int MAXIMUM_DATA_SIZE = MAXIMUM_SEGMENT_SIZE - Header.HEADER_SIZE;
	
	private Header header;
	private DatagramPacket packet;
	private long transmissionTime;
	
	
	public long getTransmissionTime() {
		return transmissionTime;
	}

	public void setTransmissionTime(long transmissionTime) {
		this.transmissionTime = transmissionTime;
	}

	private void putHeader(byte[] data, int offset) {
		if (data.length >= Header.HEADER_SIZE && data.length >= offset + Header.HEADER_SIZE) {
			if (this.header.getFlags() != null) {
				data[offset] = this.header.getFlags().getFlag();
			} else {
				data[offset] = 0;
			}

		
			if (this.header.getSegmentSequenceNumber() != null) {
				int number = this.header.getSegmentSequenceNumber().getInteger();
				byte rightByte = (byte) (number & 0xff);
				byte middleRightByte = (byte) ((number >> 8) & 0xff);
				byte middleLeftByte = (byte) ((number >> 16) & 0xff);
				byte leftByte = (byte) ((number >> 24) & 0xff);				
				
				data[offset + 1] = leftByte;
				data[offset + 2] = middleLeftByte;
				data[offset + 3] = middleRightByte;
				data[offset + 4] = rightByte;
			} else {
				this.setSequenceNumber(new SequenceNumber(0));
				data[offset + 1] = 0;
				data[offset + 2] = 0;
				data[offset + 3] = 0;
				data[offset + 4] = 0;
			}
			
			if (this.header.getAcknowledgmentNumber() != null) {
				int number = this.header.getAcknowledgmentNumber().getInteger();
				byte rightByte = (byte) (number & 0xff);
				byte middleRightByte = (byte) ((number >> 8) & 0xff);
				byte middleLeftByte = (byte) ((number >> 16) & 0xff);
				byte leftByte = (byte) ((number >> 24) & 0xff);				
				
				data[offset + 5] = leftByte;
				data[offset + 6] = middleLeftByte;
				data[offset + 7] = middleRightByte;
				data[offset + 8] = rightByte;
			} else {
				this.setAcknowledgmentNumber(new SequenceNumber(0));
				data[offset + 5] = 0;
				data[offset + 6] = 0;
				data[offset + 7] = 0;
				data[offset + 8] = 0;
			}

		} else {			
			throw new HeaderDontFitException();
		}
	}
	
	public Segment(DatagramPacket packet){
		this.packet = packet;
		this.header = new Header(packet);
	}
	
	public Segment(Header header, DatagramPacket packet) {
		super();
		this.header = header;
		this.packet = packet;
		
		this.putHeader(packet.getData(), packet.getOffset());		
	}

	//SEG.UP  - segment urgent pointer
	// SEG.PRC - segment precedence value
	public Header getHeader() {
		return this.header;
	}
	public SequenceNumber getSequenceNumber() {
		return header.getSegmentSequenceNumber();
	}
	public void setSequenceNumber(SequenceNumber sequenceNumber) {
		this.header.setSegmentSequenceNumber(sequenceNumber);
		this.putHeader(packet.getData(), packet.getOffset());		
	}
	public SequenceNumber getAcknowledgmentNumber() {
		return header.getAcknowledgmentNumber();
	}
	public void setAcknowledgmentNumber(SequenceNumber acknowledgmentNumber) {
		this.header.setAcknowledgmentNumber(acknowledgmentNumber);
		this.putHeader(packet.getData(), packet.getOffset());		
	}
	public int getLength() {
		return (this.packet.getLength()-this.packet.getOffset()) - Header.HEADER_SIZE;
	}
	@SuppressWarnings("unused")
	private void setSegmentLength(int segmentLength) {
		this.packet.setLength(segmentLength + Header.HEADER_SIZE);
	}
//	public int getSegmentWindow() {
//		return segmentWindow;
//	}
//	public void setSegmentWindow(int segmentWindow) {
//		this.segmentWindow = segmentWindow;
//	}
	public DatagramPacket getPacket() {
		return packet;
	}
	public void setPacket(DatagramPacket packet) {
		this.packet = packet;
		this.putHeader(packet.getData(), packet.getOffset());		
	}
	
	
	
	
	
	
}
