package comunicacao.sender;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.SocketTimeoutException;
import java.util.PriorityQueue;

import comunicacao.Socket;
import comunicacao.data.segment.Header;
import comunicacao.data.segment.HeaderFlags;
import comunicacao.data.segment.Segment;
import comunicacao.data.segment.SequenceNumber;
import comunicacao.util.ConnectionState;
import comunicacao.util.ReceiveNextComparator;

public class ReceiverThread implements Runnable {
	private Socket socket;

	private byte[] receiveBuffer;
	private int offset;
	private int length;
	private Object receivingMonitor;

	private Object downloadRateMonitor;

	private PriorityQueue<Segment> heap;

	private int totalAmountOfData;
	private int amountOfDataReceived;

	private long lastTime;
	private int downloadedData;
	private double downloadRate;

	private SequenceNumber fynAckSeqNum;

	private boolean alive;

	public ReceiverThread(Socket socket) {
		this.socket = socket;
		this.receiveBuffer = null;

		this.offset = 0;
		this.length = 0;
		this.receivingMonitor = new Object();

		this.downloadRateMonitor = new Object();

		this.heap = new PriorityQueue<Segment>(20, new ReceiveNextComparator(
				socket));

		this.totalAmountOfData = length;
		this.amountOfDataReceived = 0;

		this.lastTime = System.currentTimeMillis();
		this.downloadedData = 0;
		this.downloadRate = 0;

		this.fynAckSeqNum = null;

		this.alive = true;
	}

	public void run() {

		while (this.isAlive()) {
			DatagramPacket p = new DatagramPacket(
					new byte[Segment.MAXIMUM_DATA_SIZE + Header.HEADER_SIZE],
					Segment.MAXIMUM_DATA_SIZE + Header.HEADER_SIZE);
			try {
				try {
					this.socket.getSocket().receive(p);
				} catch (SocketTimeoutException e) {
					this.socket.getSender().setAlive(false);
					this.setAlive(false);
					this.socket.setState(ConnectionState.CLOSED);
				}
				if (this.isAlive()) {
					Segment segment = new Segment(p);
					Header header = segment.getHeader();

					synchronized (this.socket.getStateMonitor()) {

						switch (this.socket.getState()) {
						case OPENING:
							if (header.getFlags().isSyn()) {
								if (!header.getFlags().isAck()) { // Received a
																	// SYN,
																	// but not a
																	// SYNACK
									/*------------------STARTING RECEIVING VARIABLES--------------------*/
									this.socket.setReceiveInitialSeqNum(segment
											.getSequenceNumber());
									this.socket
											.setReceiveNextSeqNum(new SequenceNumber(
													segment.getSequenceNumber()
															.getInteger() + 1));
									this.socket
											.setRemoteAddress(p.getAddress());
									this.socket.setRemotePort(p.getPort());
									/*---------------------STARTING SENDING VARIABLES-------------------*/
									this.socket
											.setSendInitialSeqNum(new SequenceNumber(
													this.socket.getRandom()
															.nextInt()));
									this.socket
											.setSendNextSeqNum(new SequenceNumber(
													this.socket
															.getSendInitialSeqNum()
															.getInteger() + 1));
									this.socket
											.setSendUnackSeqNum(new SequenceNumber(
													this.socket
															.getSendInitialSeqNum()
															.getInteger()));

									/*--------------SENDING SYNACK-------------*/
									Header synAck = new Header(
											this.socket.getSendInitialSeqNum(),
											this.socket.getReceiveNextSeqNum()
													.clone(), new HeaderFlags(
													true, false, false, true,
													false, false, 0));
									this.socket.getSender().submit(synAck);
								} else { // Received a SYNACK
									/*------------------STARTING RECEIVING VARIABLES--------------------*/
									this.socket.setReceiveInitialSeqNum(segment
											.getSequenceNumber());
									this.socket
											.setReceiveNextSeqNum(new SequenceNumber(
													segment.getSequenceNumber()
															.getInteger() + 1));
									this.socket
											.setRemoteAddress(p.getAddress());
									this.socket.setRemotePort(p.getPort());

									/*--------------SENDING ACK---------------*/
									Header ack = new Header(this.socket
											.getSendNextSeqNum().clone(),
											this.socket.getReceiveNextSeqNum(),
											new HeaderFlags(false, false,
													false, true, false, false,
													0));

									this.socket.getSender().acknowledge(
											segment.getAcknowledgmentNumber(),
											false);
									if (this.socket.getSendUnackSeqNum()
											.getInteger() > this.socket
											.getSendInitialSeqNum()
											.getInteger()) {
										this.socket
												.setState(ConnectionState.OPEN);
									}
									this.socket.getSender().submit(ack);

								}

							}
							if (header.getFlags().isAck()) {// Received an ACK,
															// so,
															// the connection is
															// opened
								/*------------------CHANGING STATE------------*/
								this.socket.getSender().acknowledge(
										segment.getAcknowledgmentNumber(),
										false);
								if (this.socket.getSendUnackSeqNum()
										.getInteger() > this.socket
										.getSendInitialSeqNum().getInteger()) {
									this.socket.setState(ConnectionState.OPEN);
								}
							}
							// TODO Eu acho que ele simplesmente descarta todos
							// esses 3 casos...
							if (header.getFlags().isRst()) {

							}
							if (header.getFlags().isFyn()) {

							}
							if (header.getFlags().isFile()) {

							}
							break;
						case OPEN:

							if (header.getFlags().isSyn()) {
								Header ack = new Header(this.socket
										.getSendNextSeqNum().clone(),
										this.socket.getReceiveNextSeqNum(),
										new HeaderFlags(false, false, false,
												true, false, false, 0));
								this.socket.getSender().submit(ack);
							}
							if (header.getFlags().isAck()) {
								this.ackReceived(segment
										.getAcknowledgmentNumber());
							}
							if (header.getFlags().isRst()) {

								// TODO nao sei oq fazer aki...
							}
							if (header.getFlags().isFyn()) {

								/*------------------SENDING FYNACK--------------------*/
								// TODO verify ackNum
								Header fynAck = new Header(this.socket
										.getSendNextSeqNum().clone(),
										this.socket.getReceiveNextSeqNum(),
										new HeaderFlags(false, true, false,
												true, false, false, 0));
								this.fynAckSeqNum = fynAck
										.getSegmentSequenceNumber();
								this.socket.getSender().submit(fynAck);
								this.socket.setState(ConnectionState.CLOSING);
							}
							if (header.getFlags().isFile()) {

								// TODO The file must be treated
								this.dataReceived(segment);
							}
							break;
						case CLOSING:
							if (header.getFlags().isSyn()) {

							}
							if (header.getFlags().isFyn()) {// Received FYN,
								if (!header.getFlags().isAck()) { // but not a
																	// FYNACK
									/*-------------SENDING FYNACK-----------------*/
									Header fynAck2 = new Header(this.socket
											.getSendNextSeqNum().clone(),
											segment.getSequenceNumber(),
											new HeaderFlags(false, true, false,
													true, false, false, 0));
									this.fynAckSeqNum = fynAck2
											.getSegmentSequenceNumber();
									this.socket.getSender().submit(fynAck2);
								} else {// Received FYNACK
									/*------------SENDING ACK------------------*/
									Header ack3 = new Header(this.socket
											.getSendNextSeqNum().clone(),
											segment.getSequenceNumber(),
											new HeaderFlags(false, false,
													false, true, false, false,
													0));
									Segment seg = new Segment(
											ack3,
											new DatagramPacket(
													new byte[Header.HEADER_SIZE],
													0,
													Header.HEADER_SIZE,
													this.socket
															.getRemoteAddress(),
													this.socket.getRemotePort()));

									this.socket.getSocket().send(
											seg.getPacket());
									// this.socket.getSender().submit(ack3);

									this.ackReceived(segment
											.getAcknowledgmentNumber());

									this.socket.getSender().setAlive(false);
									this.socket.getSocket().setSoTimeout(30000);

								}

							} else if (header.getFlags().isAck()) {// Received
																	// ACK, so
																	// FYN
								// already sent
								// Do nothing, keep
								// waiting for FYNACK
								// TODO tem que verificar de alguma forma se
								// esse
								// ACK eh referente a resposta ao FYNACK, pois
								// se
								// for, o estado deve ser mudado para CLOSED

								if (segment.getAcknowledgmentNumber().equals(
										this.fynAckSeqNum)) { // That's
																// the
																// ACK
																// for
																// the
																// FYNACK
									this.ackReceived(segment
											.getAcknowledgmentNumber());
									this.socket.getSender().setAlive(false);
									this.setAlive(false);
									this.socket
											.setState(ConnectionState.CLOSED);

								}

							}
							if (header.getFlags().isRst()) {
								// TODO =S
							}

							if (header.getFlags().isFile()) {

							}
							break;
						case CLOSED:// Its closed
							if (header.getFlags().isSyn()) {

							}
							if (header.getFlags().isAck()) {

							}
							if (header.getFlags().isRst()) {

							}
							if (header.getFlags().isFyn()) {

							}
							if (header.getFlags().isFile()) {

							}
							break;
						}

					}
				}

			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	public boolean checkSequenceNumber(Segment seg) {
		boolean acceptable = false;
		if (seg.getLength() == 0) {
			if (this.socket.getReceiveWindow() == 0) {
				if (seg.getSequenceNumber().equals(
						this.socket.getReceiveNextSeqNum())) {
					acceptable = true;
				}
			} else {
				if (this.socket.getReceiveNextSeqNum().getInteger() <= seg
						.getSequenceNumber().getInteger()
						&& seg.getSequenceNumber().getInteger() < (this.socket
								.getReceiveNextSeqNum().getInteger() + this.socket
								.getReceiveWindow())) {
					acceptable = true;
				}
			}
		} else {
			if (this.socket.getReceiveWindow() == 0) {
				acceptable = false;
			} else {
				if ((this.socket.getReceiveNextSeqNum().getInteger() <= seg
						.getSequenceNumber().getInteger() && seg
						.getSequenceNumber().getInteger() < (this.socket
						.getReceiveNextSeqNum().getInteger() + this.socket
						.getReceiveWindow()))
						|| this.socket.getReceiveNextSeqNum().getInteger() <= seg
								.getSequenceNumber().getInteger()
								+ seg.getLength()
						&& seg.getSequenceNumber().getInteger()
								+ seg.getLength() < this.socket
								.getReceiveNextSeqNum().getInteger()
								+ this.socket.getReceiveWindow()) {
					acceptable = true;
				}
			}
		}
		return acceptable;
	}

	public double getDownloadRate() {

		long elapsedTime = System.currentTimeMillis() - this.lastTime;
		this.lastTime = System.currentTimeMillis();
		synchronized (downloadRateMonitor) {
			this.downloadRate = (double) this.downloadedData / elapsedTime;
			this.downloadedData = 0;
		}
		return this.downloadRate;
	}

	public void ackReceived(SequenceNumber ackSeqNum) {
		boolean acceptable = false;
//		System.out.println("Treating ack " + ackSeqNum.getInteger());
		synchronized (this.socket.getSeqNumMonitor()) {
			if (ackSeqNum.getInteger() >= this.socket.getSendUnackSeqNum()
					.getInteger()
					&& ackSeqNum.getInteger() <= this.socket
							.getSendNextSeqNum().getInteger()) {

				acceptable = true;
//				System.out.println("IF1");

			} else if ((ackSeqNum.getInteger() >= this.socket
					.getSendUnackSeqNum().getInteger() || ackSeqNum
					.getInteger() <= this.socket.getSendNextSeqNum()
					.getInteger())
					&& (this.socket.getSendNextSeqNum().getInteger() < this.socket
							.getSendUnackSeqNum().getInteger())) {
//				System.out.println("IF2");
				acceptable = true;

			} else {
//				System.out.println("IF3");
				acceptable = false;
			}

			if (acceptable) {
				this.socket.getSender().acknowledge(ackSeqNum, true);
			}
		}
	}

	public void dataReceived(Segment segment) {

		SequenceNumber seqNum = segment.getSequenceNumber();
		boolean aux = (seqNum.getInteger() >= this.socket.getReceiveNextSeqNum().getInteger());
		boolean aux2 = seqNum.getInteger() < this.socket.getReceiveWindow()+ this.socket.getReceiveNextSeqNum().getInteger();
		boolean aux3 = (this.socket.getReceiveWindow()+ this.socket.getReceiveNextSeqNum().getInteger()) < this.socket.getReceiveNextSeqNum().getInteger();
		boolean aux5 = seqNum.getInteger() < (this.socket.getReceiveNextSeqNum().getInteger() + this.socket.getReceiveWindow());
		
		if ((aux)
				|| (( aux3 && (aux || aux5)))) {
			if (this.socket.getReceiveNextSeqNum().getInteger()
					+ this.socket.getReceiveWindow() > seqNum.getInteger()
					+ segment.getLength() || (aux3 && ( seqNum.getInteger()
							+ segment.getLength() > this.socket.getReceiveNextSeqNum().getInteger() || seqNum.getInteger()
							+ segment.getLength() <  this.socket.getReceiveNextSeqNum().getInteger()
							+ this.socket.getReceiveWindow() )) ) {
				// Segmento dentro da janela
				// Processa dados se o numero de sequencia for igual ao
				// receiveNext, bufferiza se nao for

				if (seqNum.getInteger() == this.socket.getReceiveNextSeqNum()
						.getInteger()) {
					synchronized (receivingMonitor) {
						if (this.receiveBuffer != null) {

							synchronized (downloadRateMonitor) {

								for (int i = 0, j = Header.HEADER_SIZE; i < segment
										.getLength(); i++, offset++, j++, this.amountOfDataReceived++, this.downloadedData++) {
//									System.out.println("Segment length: "
//											+ segment.getLength() + " offset "
//											+ this.offset + " i " + i + " j "
//											+ j);
									this.receiveBuffer[offset] = segment
											.getPacket().getData()[j];
								}
							}
//							System.out.println("Amount of data "
//									+ this.amountOfDataReceived + " length: "
//									+ this.length);
							this.socket.getReceiveNextSeqNum().add(
									segment.getLength());
							if (this.amountOfDataReceived == this.length) {
								System.out.println("Chegou tudo");

								receivingMonitor.notify();
								this.amountOfDataReceived = 0;
								this.receiveBuffer = null;
								this.offset = 0;
								this.length = 0;

							} else {

								while (this.heap.peek() != null
										&& this.heap
												.peek()
												.getSequenceNumber()
												.equals(this.socket
														.getReceiveNextSeqNum())) {
									Segment segmentAux = this.heap.remove();
									this.socket.getReceiveNextSeqNum().add(
											segmentAux.getLength());
									synchronized (downloadRateMonitor) {
										for (int i = 0, j = Header.HEADER_SIZE; i < segmentAux
												.getLength(); i++, offset++, j++, this.amountOfDataReceived++, this.downloadedData++) {
											this.receiveBuffer[offset] = segmentAux
													.getPacket().getData()[j];
										}
									}
								}
								if (this.amountOfDataReceived == this.length) {
									synchronized (receivingMonitor) {
										this.amountOfDataReceived = 0;
										receivingMonitor.notify();
										this.receiveBuffer = null;
										this.offset = 0;
										this.length = 0;
									}
								}
							}
						} else {
							this.heap.add(segment);
						}
					}
				} else {
					this.heap.add(segment);
				}

			}
			// manda ack
			Header header = new Header(this.socket.getSendNextSeqNum().clone(),
					this.socket.getReceiveNextSeqNum().clone(),
					new HeaderFlags(false, false, false, true, false, false, 0));
			this.socket.getSender().submit(header);

		}

	}

	public void read(byte[] data, int offset, int length) {
		synchronized (receivingMonitor) {

			this.receiveBuffer = data;
			this.offset = offset;
			this.length = length;
			try {
				this.receivingMonitor.wait();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		}

	}

	public synchronized boolean isAlive() {
		return alive;
	}

	public synchronized void setAlive(boolean alive) {
		this.alive = alive;
	}

}