package comunicacao.sender;

import java.io.IOException;
import java.net.DatagramPacket;
import java.util.LinkedList;
import java.util.concurrent.LinkedBlockingQueue;

import comunicacao.Socket;
import comunicacao.data.SegmentsConstructor;
import comunicacao.data.segment.Header;
import comunicacao.data.segment.Segment;
import comunicacao.data.segment.SequenceNumber;
import comunicacao.util.ConnectionState;

public class SendQueue implements Runnable {
	private Socket socket;

	private LinkedList<Header> controlSendQueue;
	private LinkedBlockingQueue<Segment> sendingQueue;
	private Object sendingQueueAddMonitor;
	private RetransmitQueue retransmissionQueue;

	private boolean alive;
	private Object monitor;

	private Object sendRequestMonitor;
	private int length;

	private int bytesAcked;
	
	
	private int amountOfDataAcked;
	private long lastTime;
	private Object upLoadRateMonitor;
	private double uploadRate;

	public Object getMonitor() {
		return monitor;
	}

	@SuppressWarnings("unused")
	private void setMonitor(Object monitor) {
		this.monitor = monitor;
	}

	public SendQueue(Socket socket) {
		this.socket = socket;

		this.controlSendQueue = new LinkedList<Header>();
		this.retransmissionQueue = new RetransmitQueue(socket);
		this.sendingQueue = new LinkedBlockingQueue<Segment>();
		this.sendingQueueAddMonitor = new Object();

		this.alive = true;
		this.monitor = new Object();
		this.bytesAcked = 0;
		
		this.amountOfDataAcked = 0;
		lastTime = System.currentTimeMillis();
		upLoadRateMonitor = new Object();
	}

	public void issueControl(Header control) {
		synchronized (monitor) {
			this.controlSendQueue.add(control);
			monitor.notify();
		}
	}

	public synchronized boolean isAlive() {
		return alive;
	}

	public synchronized void setAlive(boolean alive) {
		this.alive = alive;
		synchronized (this.monitor) {
			this.monitor.notify();
		}

		this.retransmissionQueue.getExecutor().shutdownNow();
	}

	@Override
	public void run() {
		while (this.isAlive()) {
			try {
				synchronized (monitor) {
					if (!controlSendQueue.isEmpty()) {
						Header next = controlSendQueue.peek();
						
						if (next.isFyn()) {
							if (this.sendingQueue.isEmpty()) {
								Segment fyn = new Segment(new Header(this.socket.getSendNextSeqNum(), this.socket
										.getReceiveNextSeqNum().clone(), next.getFlags()), new DatagramPacket(
										new byte[Header.HEADER_SIZE], Header.HEADER_SIZE,
										this.socket.getRemoteAddress(), this.socket.getRemotePort()));
								this.retransmissionQueue.issue(fyn);
								this.controlSendQueue.remove();
							} else {
								this.sendSegments();
							}
						}

						else if (next.isRST() || next.isSYN()) {
							SequenceNumber seq = next.getSegmentSequenceNumber();
							SequenceNumber ack = next.getAcknowledgmentNumber();
							if (seq == null) {
								seq = this.socket.getSendNextSeqNum();
							}
							if (ack == null) {
								ack = this.socket.getReceiveNextSeqNum();
							}

							Segment rstSyn = new Segment(new Header(seq, ack, next.getFlags()), new DatagramPacket(
									new byte[Header.HEADER_SIZE], Header.HEADER_SIZE, this.socket.getRemoteAddress(),
									this.socket.getRemotePort()));

							if (next.isSYN()) {

								this.retransmissionQueue.issue(rstSyn);
							} else {
								this.socket.getSocket().send(rstSyn.getPacket());
							}
							this.controlSendQueue.remove();

						} else if (next.isACK()) {
							if (this.sendingQueue.isEmpty()) {
								SequenceNumber seq = next.getSegmentSequenceNumber();
								SequenceNumber ackNum = this.socket.getReceiveNextSeqNum();
								if (seq == null) {
									seq = this.socket.getSendNextSeqNum();
								}
								// Check if getReceiveNext is equal or less then
								// the ack num, if it is, just ignore this one
								Segment ack = new Segment(new Header(seq, ackNum, next.getFlags()), new DatagramPacket(
										new byte[Header.HEADER_SIZE], Header.HEADER_SIZE,
										this.socket.getRemoteAddress(), this.socket.getRemotePort()));
								this.controlSendQueue.remove();
								this.socket.getSocket().send(ack.getPacket());

							} else {
								// Piggyback
								this.sendSegments();
							}
						}

					} else if (!sendingQueue.isEmpty()) {
						// piggyback
						this.sendSegments();

					} else {
						// Nothing to send, go to sleep.
						try {
							monitor.wait(30000);
						} catch (InterruptedException e) {
							// e.printStackTrace();
						}
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
			}

		}
	}

	public void sendSegments() throws IOException {
		Segment seg;
		boolean transmit;
		synchronized (this.socket.getSeqNumMonitor()) {
			int una = this.socket.getSendUnackSeqNum().getInteger();
			int window = this.socket.getSendWindow();
			do {
				seg = this.sendingQueue.peek();
				if (seg != null) {
					if (seg.getSequenceNumber().getInteger() < una + window && seg.getSequenceNumber().getInteger() >= una) {
						transmit = true;
					} else if (una + window < una
							&& (seg.getSequenceNumber().getInteger() >= una || seg.getSequenceNumber().getInteger() < una
									+ window)) {
						transmit = true;
					} else {
						transmit = false;
					}
//					System.out.println(/*"Send queue: " + this.sendingQueue.size() + " retransmit queue: " + this.retransmissionQueue.getRetransmissionQueue().size() + "\n*/"transmit " + transmit  + " Send header: " + seg.getSequenceNumber().getInteger() + " Window+una " + (window + una)   +" Una "  + una);
					if (transmit) {
						this.sendingQueue.remove();
						// TODO piggybacking
						this.retransmissionQueue.issue(seg);
					} else {
						Thread.yield();
					}
				} else {
					transmit = false;
				}
			} while (seg != null && transmit);
		}

	}

	public void submit(Header header) {
		synchronized (monitor) {
			this.controlSendQueue.add(header);
			monitor.notify();
		}

	}

	public void submit(Segment segment) {
		synchronized (sendingQueueAddMonitor) {
			this.sendingQueue.add(segment);
			synchronized (monitor) {
				monitor.notify();
			}
		}
	}

	public void submit(byte[] data, int offset, int length, Object sendingMonitor) {
		this.length = length;
		this.sendRequestMonitor = sendingMonitor;
		new SegmentsConstructor(data, length, this.socket.getRemoteAddress(), this.socket.getRemotePort(), this.socket);

	}

	public double getUploadRate() {
		long elapsedTime = System.currentTimeMillis() - this.lastTime;
		this.lastTime = System.currentTimeMillis();
		synchronized (upLoadRateMonitor) {
			this.uploadRate = (double) this.amountOfDataAcked / elapsedTime;
			this.amountOfDataAcked = 0;
		}
		return this.uploadRate;
	}

	public void acknowledge(SequenceNumber ackSeqNum, boolean isData) {
		// TODO Implement acknowledge(SequenceNumber)
		synchronized (this.socket.getSeqNumMonitor()) {
			if (this.socket.getSendUnackSeqNum().getInteger() < this.socket.getSendNextSeqNum().getInteger()) {
				// Common case
//				System.out.println("IF4");
				if (this.socket.getSendUnackSeqNum().getInteger() <= ackSeqNum.getInteger()) {
//					System.out.println("IF5");
					if (isData) {
						bytesAcked += ackSeqNum.getInteger() - this.socket.getSendUnackSeqNum().getInteger();
						amountOfDataAcked +=  ackSeqNum.getInteger() - this.socket.getSendUnackSeqNum().getInteger();
					}
					this.socket.getSendUnackSeqNum().advanceToReach(ackSeqNum);
//					System.out.println("New send una " + this.socket.getSendUnackSeqNum().getInteger());
					this.retransmissionQueue.acknowledge();

				}
			} else if (this.socket.getSendUnackSeqNum().getInteger() > this.socket.getSendNextSeqNum().getInteger()) {
				boolean aux = this.socket.getSendUnackSeqNum().getInteger() <= ackSeqNum.getInteger();
				boolean aux2 = ackSeqNum.getInteger() < this.socket.getSendNextSeqNum().getInteger();
//				System.out.println("IF6");
				if (aux || aux2) {
//					System.out.println("IF7");
					if (isData) {
						if (aux) {
							bytesAcked += ackSeqNum.getInteger() - this.socket.getSendUnackSeqNum().getInteger();
							amountOfDataAcked +=  ackSeqNum.getInteger() - this.socket.getSendUnackSeqNum().getInteger();
						} else if (aux2) {
							bytesAcked += ackSeqNum.getInteger() - Integer.MIN_VALUE + 1;
							amountOfDataAcked +=   ackSeqNum.getInteger() - Integer.MIN_VALUE + 1;
							bytesAcked += Integer.MAX_VALUE - this.socket.getSendUnackSeqNum().getInteger();
							amountOfDataAcked +=  Integer.MAX_VALUE - this.socket.getSendUnackSeqNum().getInteger();
						}
					}
					this.socket.getSendUnackSeqNum().advanceToReach(ackSeqNum);
					this.retransmissionQueue.acknowledge();
				}
			}
		}
		
//		System.out.println(this.bytesAcked + " bytes were acked, total of " + this.length);
		if (this.bytesAcked == this.length && this.sendRequestMonitor != null) {
			synchronized (this.sendRequestMonitor) {
				this.bytesAcked = 0;
				sendRequestMonitor.notify();
				this.sendRequestMonitor = null;
			}
		}
	}

	public int getFlightSize() {
		// TODO Implement getFlightSize()
		return 0;
	}

	public synchronized Object getSendingQueueAddMonitor() {
		return sendingQueueAddMonitor;
	}

	public synchronized void setSendingQueueAddMonitor(Object sendingQueueAddMonitor) {
		this.sendingQueueAddMonitor = sendingQueueAddMonitor;
	}

}
