package core;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import userManagement.User;

import comunicacao.ISocket;
import comunicacao.Socket;
import comunicacao.exception.NotConnectedException;

public class ClienteComunication extends ProgramServer{

	private String ip = "localhost";
	ISocket socket;
	private ExecutorService serverThreads;
	private ArrayList<SendingCommand> threads;
	public ClienteComunication(){
		super();
		serverThreads = Executors.newCachedThreadPool();
		threads = new ArrayList<SendingCommand>();
	}
	
	public boolean registerUser(String name, String email, String password)
			throws IOException, NotConnectedException {
		socket = new Socket(InetAddress.getByName(ip),2000);
		//int newPort = (new BufferedReader(new InputStreamReader(socket.getInputStream()))).read();
		//Socket socket2 = new Socket("localhost",newPort);
		socket.sendObject("registerUser");
		socket.sendObject(name);
		socket.sendObject(email);
		socket.sendObject(password);
		boolean ok = false;
		try {
			ok = (Boolean) socket.readObject();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		//socket2.close();
		socket.close();
		return ok;
		//return super.registerUser(name, email, password);
	}

	public User verifyUser(String email, String password)
			throws ClassNotFoundException, IOException, NotConnectedException {
		socket = new Socket(InetAddress.getByName(ip),2000);
		//int newPort = (new BufferedReader(new InputStreamReader(socket.getInputStream()))).read();
		//Socket socket2 = new Socket("localhost",newPort);
		socket.sendObject("verifyUser");
		socket.sendObject(email);
		socket.sendObject(password);
		if((Boolean) socket.readObject()){
			User user = (User) socket.readObject();
			//socket.close();
			return user;
		}
		//User user = new User(socketStringInput.readLine(),socketStringInput.readLine(),socketStringInput.readLine());
		//socket.close(); se for null fecha a conexo.
		//socket2.close();
		socket.close();
		return null;
		//return super.verifyUser(email, password);
	}

	public void putFile(User user, String address, File file) {
		try {
			//socket = new Socket("localhost",2000);
			socket.sendObject("putFile");
			int port = (Integer) socket.readObject();
			SendingCommand sendingCommand = new SendingCommand(ip, port,"putFile",user,address,file);
			threads.add(sendingCommand);
			serverThreads.submit(sendingCommand);
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotConnectedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
	}

	public void putDirectory(User user, String address) {
		try {
			//socket = new Socket("localhost",2000);
			socket.sendObject("putDirectory");
			int port = (Integer) socket.readObject();
			SendingCommand sendingCommand = new SendingCommand(ip, port,"putDirectory",user,address);
			serverThreads.submit(sendingCommand);
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotConnectedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void delete(User user, String address) {
		try {
			socket.sendObject("deletar");
			int port = (Integer) socket.readObject();
			SendingCommand sendingCommand = new SendingCommand(ip, port,"deletar",user,address);
			serverThreads.submit(sendingCommand);
		} catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (NotConnectedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
	}
	
	public ArrayList<String> synchronizing(User user, File file,String address){
		ArrayList<String> files = new ArrayList<String>();
		try {
			socket.sendObject("synchronizing");
			socket.sendFile(file);
			System.out.println("O arquivo enviando para synchronizing eh "+file.getName()+ file.canRead());
			String command = (String) socket.readObject();
			while(!command.equals("ok")){
				if(command.equals("send")){
					String fileName = (String) socket.readObject();
					File newFile = new File(address+fileName);
					int port = (Integer) socket.readObject();
					SendingCommand sendingCommand = new SendingCommand(ip, port,"send",user,fileName,newFile);
					threads.add(sendingCommand);
					serverThreads.submit(sendingCommand);
				}
				else{
					String fileName = (String) socket.readObject();
					fileName = address+fileName;
					int port = (Integer) socket.readObject();
					SendingCommand sendingCommand = new SendingCommand(ip, port,"receive",user,fileName);
					threads.add(sendingCommand);
					serverThreads.submit(sendingCommand);
					files.add(fileName);
				}
				command = (String) socket.readObject();
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotConnectedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return files;
	}
	

	@Override
	public void desconect() {
		try {
			socket.sendObject("desconect");
			socket.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotConnectedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public ArrayList<InformationCliente> getInformationClientes(){
		ArrayList<InformationCliente> informationClientes = new ArrayList<InformationCliente>();
		for (int i = 0; i < threads.size(); i++) {
			if(threads.get(i).isActive()){
				informationClientes.add(threads.get(i).getInformationCliente());
			}
			else{
				threads.remove(i--);
			}
		}
		return informationClientes;
	}

	@Override
	public void setIp(String ip) {
		this.ip = ip;
	}
	
}
