package userManagement;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.Hashtable;

public class UserManager implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private static UserManager manager;
	public static final String userDatabaseFilePathname = "UserDatabase.yamdata";

	private Hashtable<String, User> userDatabase;
	private Hashtable<InetSocketAddress, User> loggedUsers;
	
	private ProgramServer programServer;

	private UserManager() {
		this.userDatabase = new Hashtable<String, User>();
		this.loggedUsers = new Hashtable<InetSocketAddress, User>();
	}

	public void addUser(User user) throws UserAlreadyExistentException{
		if(user != null){
			if(!userDatabase.containsKey(user.getEmail())){
				userDatabase.put(user.getEmail(), user);
			} else {
				throw new UserAlreadyExistentException(user.getEmail());
			}			
		}		
	}

	public User removeUser(String email) throws UserNotFoundException{
		User user = userDatabase.remove(email);

		if(user == null) {
			throw new UserNotFoundException(email);
		}
		return user;

	}

	public User removeUser(User user){
		userDatabase.remove(user.getEmail());
		return user;

	}


	public void updateUser(User user) throws UserNotFoundException{
		User userAux = null;

		if(user != null){
			userAux = userDatabase.get(user.getEmail());
		}	

		if(userAux != null){
			userAux.setCurrentIP(user.getCurrentIP());
			userAux.setCurrentPort(user.getCurrentPort());
			userAux.setName(user.getName());
			userAux.setPassword(user.getPassword());
		} else {
			throw new UserNotFoundException(user.getEmail());			
		}

	}

	public void updateUserEmail(String oldEmail, String newEmail) throws UserNotFoundException{
		User user = userDatabase.get(oldEmail);
		
		if(user != null){
			user.setEmail(newEmail);
		} else {
			throw new UserNotFoundException(oldEmail);
		}
		
	}

	public User login(String email, String senha, InetAddress ip, int port){
		User user = this.programServer.verifyUser(email, senha);		
		if(user != null){			
			user.setCurrentIP(ip);
			user.setCurrentPort(port);
			InetSocketAddress ipPort = new InetSocketAddress(ip, port);
			loggedUsers.put(ipPort, user);
		}
		return null;
		
	}

	public User logout(InetSocketAddress ipPort){		
		return loggedUsers.remove(ipPort);
	}

	public User search(String email){ 		
		return userDatabase.get(email);
	}
	
	public boolean isLoggedIn(InetSocketAddress ipPort){
		return loggedUsers.containsKey(ipPort);
	}

	public void shutdown() throws IOException {
		//Getting  persistent file
		File userDatabaseFile = new File(userDatabaseFilePathname);


		try {
			//Saving database
			ObjectOutputStream out = new ObjectOutputStream(
					new FileOutputStream(userDatabaseFile));
			out.writeObject(this);
			out.close();

		} catch (FileNotFoundException e) {
			// can't happen
		} catch (IOException e) {
			throw e;
		}
	}

	public static UserManager getInstace() {
		if (manager == null) {
			manager = new UserManager();
		}
		return manager;

	}

}
